;
; Modul, das die geraeteabhaengigen Systemaufrufe ueber
; neue SWI's bereitstellt
;
; Usage:
;          rmrun minixdrv
;
; (C) 15/11/90 Martin Patzel
;



start		EQU	.
DEBUG		EQU	1


;==============================================================================
; Modul-Header
;==============================================================================
mh_start_code	DCD	0
mh_init_code	DCD	module_init
mh_finish_code	DCD	module_exit
mh_service_call	DCD	0
mh_title_string	DCD	m_title_string
mh_help_string	DCD	m_help_string
mh_keywrd_table	DCD	0
mh_swi_chunk	DCD	0
mh_swi_handler	DCD	0
mh_swi_dec_tab	DCD	m_decode_tbl
mh_swi_dec_code	DCD	0

m_title_string	=	"MinixDriver", 0
m_help_string	=	"Minix Driver", 9, "1.00 (26 Oct 1990)", 0
m_decode_tbl	=	"Minix", 0
		=	"SystemCall", 0
		=	"EnableDriver", 0
		=	"DisableDriver", 0
		=	"EnterSVC", 0
		=	"PutC", 0
		=	"GetC", 0
		=	"DiskIO", 0
		=	"DiskParms", 0
		=	"SetVector", 0
		=	"Special", 0
		=	"GetParameters", 0
		=	"GetMemoryMap", 0
		=	"GetMemoryFreeList", 0
		=	"SetTranslator", 0
		=	"OSMode", 0
		=	"SetMemoryMap", 0
		=	0
		ALIGN


		GET	$.arm.aslib.register
		GET	$.arm.aslib.arthur
		GET	$.arm.aslib.constants
		GET	$.arm.aslib.macros
		GET	h.aosdriver
		GET	h.asconst



PageMagic	EQU	&31415927	; Magic-Nummer zur Seitensuche
StackSize	EQU	1024		; Stackgroesse fuer den Treiber
VarSize		EQU	128		; Puffergroesse fuer Pathvariablen
FNameSize	EQU	10		; max. Filenamenlaenge
NumFdDrives	EQU	2		; max. Anzahl simul. FD-Drives
NumHdDrives	EQU	8		; max. Anzahl simul. HD-Drives
NumSpDrives	EQU	1		; max. Anzahl simul. Swap-Bereiche
SectorSize	EQU	512		; Sektorgroesse in Bytes
OS_Mode		EQU	&00001000	; Os-Mode Bit im MEMC
IRQQSize	EQU	10		; 10 Interruptverschachtelungen max.
IOErrMax	EQU	2		; Max. 2 Versuche bei R/W


;==============================================================================
; module_init
;
; Initialisiert diverse Modulvariablen und -strukturen
;
; in  :	-
; out :	-
; uses: R0-R3, R5, R6, R12
;==============================================================================
module_init
		MOV	R6, R10			; Register sichern
		MOV	R5, LK

		STR	R12, last_wsp
		LDR	R2, [R12, #0]		; Pointer auf Workspace
		CMP	R2, #0			; Reinitialisierung?
		BNE	mi_got_mem		; ja!
		MOV	R0, #6			; Claim memory
		LDR	R3, =data_size
		SWI	OS_Module
		STR	R2, [R12, #0]		; WSP retten
mi_got_mem	MOV	R10, R2			; R10 = WSP

		ADRL	SP, stackend
		[	DEBUG > 0
		ADR	R0, macbuffer
		STR	R0, macbuf		; !!! fuer Debug-Makros !!!
		]

		MOV	R0, #VecSWI		; Dieser soll ersetzt werden
		ADR	R1, swi_own_entry	; dahin solls gehen
		SetVec	R0, R1, R2
		STR	R2, old_swi_vector

		ADRL	R0, ioe_queue		; muss manuell initialisiert
		STR	R0, ioe_queue_ptr	; werden
		MOV	R0, #VecIRQ		; Dieser soll ersetzt werden
		ADR	R1, irq_own_entry	; dahin solls gehen
		SetVec	R0, R1, R2
		STR	R2, old_irq_vector

		MOV	R0, #0
		MOV	R1, #0
		SWI	OS_UpdateMEMC+XBit	; MEMC-Ctrl-Reg holen
		LDR	R1, =MEMCControl
		ORR	R0, R0, R1
		STR	R0, memc_control

		ADR	R0, sys_para_block
		BL	get_para

		MOV	R10, R6
		BICS	PC, R5, #OverflowFlag


;==============================================================================
; module_exit
;
; Haengt sich aus dem SWI wieder aus.
;
; in  :	-
; out :	-
; uses: R0, R1
;==============================================================================
module_exit
		LDR	R10, [R12, #0]		; WSP initialisieren
		MOV	R0, #VecSWI
		LDR	R1, old_swi_vector
		GenBra	R0, R1
		STR	R1, [R0, #0]

		MOV	R0, #VecIRQ
		LDR	R1, old_irq_vector
		GenBra	R0, R1
		STR	R1, [R0, #0]

		BICS	PC, LK, #OverflowFlag


last_wsp	%	4		; zuletzt benutzter Workspace
old_swi_vector	%	4		; SWI-Pointer (-> Arthur)
old_irq_vector	%	4		; IRQ-Pointer (-> Arthur)
ioe_queue_ptr	%	4		; Pointer in die Interruptqueue
ioe_queue	%	IRQQSize*4*4	; Die Interruptqueue selber
		LTORG


;==============================================================================
; swi_entry
;
; Diese Routine wird aufgerufen bei einem SWI-Call, der fuer dieses
; Modul bestimmt ist. Diese Routine wird vom Arthur-Betriebssystem
; aufgerufen.
;
; in  :	R11		SWI-Nummer (0-63)
;	R12		Pointer auf WSP
; out :	V-Flag		gesetzt bei Fehlerstatus
; uses: R10,R12
;==============================================================================
swi_entry
		LDR	R10, [R12, #0]		; WSP initialisieren
		ADRL	SP, stackend
		CMP	R11, #(endofswitable-startofswitable)/4
		ADDCC	PC, PC, R11, LSL #2
		B	unknown_swi
startofswitable	B	unknown_swi		; war: swi_sys_call
		B	swi_en_drv
		B	swi_dis_drv
		B	swi_enter_svc
		B	swi_putc
		B	swi_getc
		B	swi_diskio
		B	swi_diskparms
		B	swi_set_vector
		B	swi_special
		B	swi_get_para
		B	swi_get_map
		B	swi_free_pages
		B	swi_set_trnsl
		B	swi_os_mode
		B	swi_set_map
endofswitable

unknown_swi	ADR	R0, swi_err_msg
		ORRS	PC, LK, #OverflowFlag

swi_err_msg	DCD	&1E6
		=	"Unknown Minix Driver operation", 0
		ALIGN



;==============================================================================
; swi_own_entry
;
; Diese Routine wird aufgerufen bei einem SWI-Call, falls Minix laeuft
; und das Arthur Betriebssystem ausgehaengt ist.
;
; in  :	-
; out :	-
; uses: -
;==============================================================================
swi_own_entry
		STMDB	R13!, {R9-R12,LK}	; R13 ist SSP!
		BIC	R9, LK, #&FC000003	; CC-Bits loeschen
		LDR	R12, [R9, #-4]		; SWI-Opcode holen
		ADR	R11, soe_return		; neue Returnadresse
		ORR	LK, R11, #SVCMode	; nachher im SVC bleiben
		BIC	R11, R12, #&FF000000	; SWI-Nummer nach R11
		BIC	R11, R11, #XBit
		LDR	R12, last_wsp
		LDR	R10, [R12, #0]		; WSP holen
		SUBS	R11, R11, #MinixSWIBase	; Minix-Driver-Call?
		BCC	soe_notmine
		CMP	R11, #64
		BCS	soe_notmine
		MOV	R9, #0
		STR	R9, swi_ret_mode	; Keine Umschaltung in SVC
		B	swi_entry

soe_return
		LDR	R9, [R13, #16]		; altes LK-Register holen
		MOV	R11, PC
		AND	R11, R11, #OverflowFlag	; neues V-Flag
		BIC	R9, R9, #OverflowFlag	; alte NZC-Flags/Mode
		ORR	LK, R9, R11
		LDR	R11, swi_ret_mode	; Return-Modus
		ORR	LK, LK, R11
		LDMIA	R13!, {R9-R12}
		ADD	R13, R13, #4
		MOVS	PC, LK			; Return vom SWI

soe_notmine
		LDMIA	R13!, {R9-R12,LK}
		LDR	PC, old_swi_vector


;==============================================================================
; irq_own_entry
;
; Diese Routine wird aufgerufen bei einem Interrupt.
;
; Mechanismus:
; IRQ tritt auf -> Sprung nach irq_own_entry.
; Von dort wird zur Arthur-Irq-Routine gesprungen. Diese entschluesselt
; die Ursache des Interrupts, ruft evtl. Interruptroutine auf.
; Diese legt die interruptspezifischen Daten in der Interruptqueue ab.
; Nach dem Ruecksprung zu Arthur springt dieses zu irq_queue_beg.
; Diese arbeitet nun ggf. die Queue Stueck fuer Stueck ab, und zwar nur im
; obersten Interruptlevel (keine verschachtelten Interrupts)
; in  :	-
; out :	-
; uses: -
;==============================================================================
irq_own_entry
		SUB	LK, LK, #4		; Prefetch Korrektur
		STMDB	R13!, {LK}		; LK_irq sichern

		ADR	LK, ioe_queue_beg+4	; Prefetch-Korrektur
		ORR	LK, LK, #IRQFlag:OR:FIRQFlag:OR:IRQMode
		LDR	PC, old_irq_vector

ioe_queue_beg
		STMDB	R13!, {R0-R3}		; Irq Queue untersuchen
		ADRL	R0, ioe_queue
		LDR	R3, ioe_queue_ptr
		SUBS	R2, R3, R0		; Queue leer?
		LDMLEIA	R13!, {R0-R3,PC}^	; ja!

		CMP	R2, #16			; mehr als ein Interrupt in
		BLE	ioe_last_one		; der Queue?
		LDMIA	R13!, {R0-R3,LK}
		STMDB	R13!, {R0-R4,LK}
		ADR	R0, ioe_queue_beg
		ORR	R0, R0, #IRQFlag:OR:FIRQFlag:OR:IRQMode
		STR	R0, [R13, #4*4]
		LDR	R3, ioe_queue_ptr
ioe_last_one
		LDMDB	R3!, {R0-R2,LK}		; gespeicherte Parameter holen
		STR	R3, ioe_queue_ptr	; neues Queueende sichern
		ORRS	PC, LK, #IRQFlag:OR:FIRQFlag:OR:IRQMode


;==============================================================================
; swi_en_drv
;
; nimmt weitere Initialisierungen zum Betrieb von Minix vor
;
; in  : -
; out : -
; uses: -
;==============================================================================
swi_en_drv
		STMDB	SP!, {R0-R4,LK}

		MOV	R0, #IOC_BASE		; Timer 1 initialisieren
		LDRB	R1, [R0, #IOC_IRQAMASK]
		BIC	R1, R1, #&40		; Timer 1 Interrupt aus
		STRB	R1, [R0, #IOC_IRQAMASK]
		MOV	R1, #&FF
		STRB	R1, [R0, #IOC_T1LATCHL]
		STRB	R1, [R0, #IOC_T1LATCHH]
		STRB	R1, [R0, #IOC_T1GO]

		MOV	R0, #1			; OS Mode aktivieren
		BL	swi_os_mode
		ADRL	R0, map			; Pagemapping holen
		BL	swi_get_map

		LDR	R0, act_ppg_num		; Protectlevel setzen
		LDR	R1, act_pg_size
		GetLog2	R1, R4			; fuer Berechnung der Pageno.
		ADRL	R3, map
		MOV	R2, #ProtectOS
		B	edr_prtct_tst
edr_prtct_loop	LDR	R1, [R3, R0, LSL #2]	; logische Seitennummer holen
		MOV	R1, R1, LSR R4
		BL	swi_set_trnsl
edr_prtct_tst	SUBS	R0, R0, #1
		BGE	edr_prtct_loop

		ADR	R0, edr_vdu_init	; Bildschirm initialisieren
		MOV	R1, #edr_vdu_init_end-edr_vdu_init
		SWI	OS_WriteN+XBit
		LDMIA	SP!, {R0-R4,LK}
		BICS	PC, LK, #OverflowFlag

edr_vdu_init	=	26			; reset screen
		=	15			; page mode off
		=	12			; clear screen
		=	23,16,&50,0,0,0,0,0,0,0	; define cursor movement
edr_vdu_init_end


;==============================================================================
; swi_dis_drv
;
; Macht die Initialisierungen von swi_en_drv rueckgaengig.
; Haengt auch die Interruptvektoren aus.
;
; in  : -
; out : -
; uses: -
;==============================================================================
swi_dis_drv
		STMDB	SP!, {R0-R4,LK}

		MOV	R0, #SV_ClockVec
		MOV	R1, #0			; Vektor freigeben
		BL	swi_set_vector
		MOV	R0, #SV_KeybdVec
		MOV	R1, #0			; Vektor freigeben
		BL	swi_set_vector

;		MOV	R0, #&1C		; TickerV = Clock Vektor
;		ADRL	R1, sve_clock
;		MOV	R2, #0
;		SWI	OS_Release+XBit
;
;		MOV	R0, #&10		; EventV = Event Vektor
;		ADRL	R1, sve_keybd
;		MOV	R2, #0
;		SWI	OS_Release+XBit
;
		LDR	R0, act_ppg_num		; alle Seiten freigeben
		LDR	R1, act_pg_size
		GetLog2	R1, R4
		ADRL	R3, map
		MOV	R2, #ProtectFree
		B	ddr_prtct_tst
ddr_prtct_loop	LDR	R1, [R3, R0, LSL #2]	; log. Seitenadresse holen
		MOV	R1, R1, LSR R4
		BL	swi_set_trnsl
ddr_prtct_tst	SUBS	R0, R0, #1
		BGE	ddr_prtct_loop

		MOV	R0, #0			; OS Mode deaktivieren
		BL	swi_os_mode

		ADR	R0, ddr_vdu_exit	; Bildschirm ruecksetzen
		MOV	R1, #ddr_vdu_exit_end-ddr_vdu_exit
		SWI	OS_WriteN+XBit

		MOV	R0, #&15		; Flush selected buffer
		MOV	R1, #0			; keyboard buffer
		SWI	OS_Byte+XBit

		LDMIA	SP!, {R0-R4,LK}
		BICS	PC, LK, #OverflowFlag

ddr_vdu_exit	=	26			; reset screen
		=	15			; page mode off
		=	12			; clear screen
		=	23,16,0,0,0,0,0,0,0,0	; define cursor movement
ddr_vdu_exit_end


;==============================================================================
; swi_enter_svc
;
; in  : -
; out : -
; uses: -
;==============================================================================
swi_enter_svc
		MOV	R11, #SVCMode
		STR	R11, swi_ret_mode	; Return im SVC!
		BICS	PC, LK, #OverflowFlag


;==============================================================================
; swi_putc
;
; Druckt ein Zeichen auf dem Bildschirm aus
;
; in  : ASCII-Zeichen in R0
;	R1,... evtl. weitere Parameter
;	R0 wird aufgeteilt in 4 Bytes: Byte 0 (B0) = LSB, Byte 3 (B3) = MSB
;	B3 = 0	: Druckt Zeichen in B0 mit Attribut in B1
;	B3 = 1	: Setzt Cursor auf (x,y) mit x in B0, y in B1
;	B3 = 2  : Scrollt Bildschirmfenster: B0=Anzahl Zeilen/Spalten,
;		  B1: 0=rechts, 1=links, 2=unten, 3=oben
;		  R1: Fenstergroesse/-position: B0=x, B1=y, B2=w, B3=h
;	B3 = 3	: Liefert die aktuellen Fensterausmasse in R0 (Format wie R1)
;
; out : bei B3=3 enthaelt R0 die aktuellen Fensterausmasse
; uses: -
;==============================================================================
swi_putc
		STMDB	SP!, {R0-R3,LK}

		ANDS	R3, R0, #&FF000000
		BNE	pch_nochout		; Zeichen nur ausgeben
		AND	R0, R0, #&FF
		SWI	OS_WriteC+XBit
		B	pch_end

pch_nochout	TEQ	R3, #PC_MoveTo		; Cursor neu setzen
		BNE	pch_nomove
		ADR	R1, pch_mvcmd
		STRB	R0, [R1, #1]		; x-Koordinate
		MOV	R0, R0, LSR #8
		STRB	R0, [R1, #2]		; y-Koordinate
		MOV	R0, R1			; Adresse des Kommandostrings
		MOV	R1, #3			; Laenge des Kommandostrings
		SWI	OS_WriteN+XBit
		B	pch_end
pch_mvcmd	DCB	31, 0, 0		; move(x,y)

pch_nomove	TEQ	R3, #PC_Scroll		; BSFenster scrollen
		BNE	pch_noscrl
		ADR	R0, pch_inblk
		ADR	R1, pch_outblk
		SWI	OS_ReadVduVariables+XBit; alte Fensterausmasse holen
		LDMIA	SP, {R0-R1}
		ADR	R3, pch_wincmd+1
		AND	R0, R1, #&FF
		STRB	R0, [R3, #0]		; linker Rand (lx)
		MOV	R2, R1, LSR #16		; Breite
		AND	R2, R2, #&FF
		SUB	R2, R2, #1
		ADD	R0, R0, R2
		STRB	R0, [R3, #2]		; rechter Rand (rx)
		MOV	R0, R1, LSR #8
		AND	R0, R0, #&FF
		STRB	R0, [R3, #3]		; oberer Rand (ty)
		MOV	R2, R1, LSR #24
		AND	R2, R2, #&FF
		SUB	R2, R2, #1
		ADD	R0, R0, R2
		STRB	R0, [R3, #1]		; unterer Rand (by)
		ADR	R0, pch_wincmd
		MOV	R1, #5
		SWI	OS_WriteN+XBit		; Window setzen

		LDMIA	SP, {R2}
		ADR	R0, pch_scrcmd
		MOV	R1, R2, LSR #8
		STRB	R1, [R0, #3]		; Richtung
		AND	R2, R2, #&FF
		MOV	R1, #10
pch_scrloop	SWI	OS_WriteN+XBit
		SUBS	R2, R2, #1
		BGT	pch_scrloop

		ADR	R0, pch_wincmd
		ADR	R1, pch_outblk
		LDR	R2, [R1, #0]
		STRB	R2, [R0, #1]
		LDR	R2, [R1, #4]
		STRB	R2, [R0, #2]
		LDR	R2, [R1, #8]
		STRB	R2, [R0, #3]
		LDR	R2, [R1, #12]
		STRB	R2, [R0, #4]
		MOV	R1, #5
		SWI	OS_WriteN+XBit
		B	pch_end

pch_inblk	DCD	132,133,134,135,-1	; alte Fensterausmasse holen
pch_outblk	DCD	0,0,0,0			; Ergebnisse
pch_wincmd	DCB	28,0,0,0,0		; lx, by, rx, ty
pch_scrcmd	DCB	23,7,0,0,0,0,0,0,0,0	; Scroll-Kommando

pch_noscrl	TEQ	R3, #PC_GetWindow
		BNE	pch_nogetw
		ADR	R0, pch_inblk
		ADR	R1, pch_outblk
		SWI	OS_ReadVduVariables+XBit; alte Fensterausmasse holen
		LDR	R0, [R1, #0]		; linker Fensterrand
		LDR	R2, [R1, #8]
		SUB	R2, R2, R0
		ADD	R2, R2, #1
		ORR	R0, R0, R2, LSL #16	; Breite
		LDR	R2, [R1, #12]
		ORR	R0, R0, R2, LSL #8	; oberer Rand
		LDR	R3, [R1, #4]
		SUB	R2, R3, R2
		ADD	R2, R2, #1
		ORR	R0, R0, R2, LSL #24	; Hoehe
		ADD	SP, SP, #4
		LDMIA	SP!, {R1-R3,LK}
		BICS	PC, LK, #OverflowFlag

pch_nogetw
pch_end		LDMIA	SP!, {R0-R3,LK}
		BICS	PC, LK, #OverflowFlag


;==============================================================================
; swi_getc
;
; Holt eine Taste von der Tastatur ab.
;
; in  : -
; out : R0, ASCII-Code in Bit 0-7
; uses: -
;==============================================================================
swi_getc
		STMDB	SP!, {LK}
		SWI	OS_ReadC+XBit
		LDMIA	SP!, {LK}
		BICS	PC, LK, #OverflowFlag


;==============================================================================
; swi_diskio
;
; Treiberroutine zum lesen/schreiben von Sektoren von/auf Diskette/Platte
;
; in  :	R0	0=Lesen, 1=Schreiben
;	R1	Laufwerksnummer (0=Floppy, 1=Harddisk, 2=Swap)
;	R2	Partitionnummer (00-FF)
;	R3	Anzahl der zu lesenden/schreibenden Sektoren
;	R4	logische Sektornummer des ersten Sektors
;	R5	Pufferadresse
; out :	R0	0 bzw. VC = ok, sonst Fehlercode und VS
; uses: -
;==============================================================================
swi_diskio
		STMDB	SP!, {R0-R8, LK}
;		HPrint	R1
;		SPrint	" "
;		HPrint	R3
;		SPrint	" "
;		HPrint	R4
;		SPrint	" "
;		HPrint	R5
;		NLPrint	
		CMP	R2, #0
		BLT	dio_badparm		; zu kleine Drivenummer
		MOV	R6, #0
		STR	R6, dio_err_cnt
		TEQ	R1, #Flop_Dev
		BNE	dio_hd			; nicht die Floppy

dio_fd		CMP	R2, #NumFdDrives
		BGE	dio_badparm		; zu hohe Drivenummer
		ADR	R7, dio_ffd_tbl
		LDR	R6, [R7, R2, LSL #2]	; fd holen
		CMP	R6, #0
		BNE	dio_open		; File ist schon offen!
		B	dio_open_it

dio_hd		TEQ	R1, #Hard_Dev
		BNE	dio_sp			; nicht die Harddisk
		CMP	R2, #NumHdDrives	; zu hohe Drivenummer
		BGE	dio_badparm
		ADR	R7, dio_hfd_tbl
		LDR	R6, [R7, R2, LSL #2]
		CMP	R6, #0
		BNE	dio_open		; File ist schon offen!
		B	dio_open_it

dio_sp		TEQ	R1, #Swap_Dev
		BNE	dio_badparm		; auch nicht Swap => Error
		CMP	R2, #NumSpDrives	; zu hohe Drivenummer
		BGE	dio_badparm
		ADR	R7, dio_sfd_tbl
		LDR	R6, [R7, R2, LSL #2]
		CMP	R6, #0
		BNE	dio_open		; File ist schon offen!
		B	dio_open_it

dio_try_again	LDMIA	SP,{R0-R2}
		MOV	R0, #0
		LDR	R1, [R7, R2, LSL #2]
		SWI	OS_Find+XBit		; File schliessen
		LDMIA	SP,{R0-R5}

dio_open_it	BL	make_filename
		MOV	R1, R0			; Pointer auf Filenamen
		MOV	R0, #&CF		; kein Pfad, open for update
		SWI	OS_Find+XBit
		BVS	dio_openerr		; open error!
		STR	R0, [R7, R2, LSL #2]	; Filedeskriptor -> Tabelle
		MOV	R6, R0

dio_open	MOV	R1, R6			; Filedeskriptor
		MOV	R0, #2			; Dateigroesse holen
		SWI	OS_Args+XBit
		MOVVS	R8, #0			; Dateigroesse bei Fehler = 0
		MOVVC	R8, R2

		LDMIA	SP, {R0-R5}		; Parameter nochmal holen
		MOV	R1, R6			; Filedescriptor
		MOV	R2, R5			; Memory-Startadresse
		MOV	R6, #&FFFFFFFE
		TST	R0, R6			; R0=0 oder 1 !!!
		BNE	dio_badparm
		CMP	R0, #0			; 0=lesen
		MOVEQ	R0, #3
		MOVNE	R0, #1

		MOV	R6, #SectorSize		; Anzahl sektoren
		MUL	R6, R3, R6		; in Bytes umrechnen
		MOV	R3, R6
		MOV	R6, #SectorSize		; Startsektor in absolute
		MUL	R6, R4, R6		; Byteadresse umrechnen
		MOV	R4, R6
		ADD	R6, R3, R4		; wo kommt EOF zu liegen?
		CMP	R6, R8			; ausserhalb der Datei?
		BGT	dio_eof

;		HPrint	R0
;		SPrint	" "
;		HPrint	R2
;		SPrint	" "
;		HPrint	R3
;		SPrint	" "
;		HPrint	R4
;		NLPrint	
		SWI	OS_GBPB+XBit		; schreibe/lese Daten
		BVC	dio_rw_ok
		LDR	R8, dio_err_cnt
		CMP	R8, #IOErrMax
		BGE	dio_rw_err
		ADD	R8, R8, #1
		STR	R8, dio_err_cnt
		B	dio_try_again

dio_rw_ok	MOV	R0, #&FF
		SWI	OS_Args+XBit		; gepufferte Daten schreiben

		LDMIA	SP!, {R0-R8, LK}
		MOV	R0, #0
		BICS	PC, LK, #OverflowFlag


dio_rw_err	SWI	OS_WriteS+XBit
		=	"Minix$DiskIO: fatal: "
		=	"read/write error", 13, 10, 0
		B	dio_err1
dio_eof		SWI	OS_WriteS+XBit
		=	"Minix$DiskIO: fatal: "
		=	"eof reached", 13, 10, 0
		B	dio_err1
dio_badparm	SWI	OS_WriteS+XBit
		=	"Minix$DiskIO: fatal: "
		=	"bad parameter", 13, 10, 0
		B	dio_err1
dio_openerr	SWI	OS_WriteS+XBit
		=	"Minix$DiskIO: fatal: "
		=	"can't open drive file!", 13, 10, 0
dio_err1	MOV	R0, #-1
dio_err		ADD	SP, SP, #4
		LDMIA	SP!, {R1-R7, LK}
		ORRS	PC, LK, #OverflowFlag


dio_ffd_tbl	%	NumFdDrives*4
dio_hfd_tbl	%	NumHdDrives*4
dio_sfd_tbl	%	NumSpDrives*4
dio_err_cnt	%	4



;==============================================================================
; make_filename
;
; erzeugt aus Lauferks- und Partitionnummer einen eindeutigen Namen
; Die Namen werden auf dem Pfad Minix$Floppy bzw. Minix$Harddisk bzw.
; Minix$Swap erzeugt.
;
; in  : R1	Laufwerksnummer (0=Floppy, 1=Harddisk, 2=Swap)
;	R2	Partitionnummer (00-FF)
; out : R0	Pointer auf erzeugten Filenamen
; uses: -
;==============================================================================
make_filename
		STMDB	SP!, {R1-R4, LK}
		CMP	R1, #Flop_Dev			; Floppy?
		ADDEQ	R0, PC, #mfn_flop_name-.-8	; ja
		CMP	R1, #Hard_Dev			; Harddisk?
		ADDEQ	R0, PC, #mfn_hard_name-.-8	; ja
		ADDGT	R0, PC, #mfn_swap_name-.-8	; nein -> Swap
		ADR	R1, mfn_file_path
		MOV	R2, #VarSize-FNameSize-1
		MOV	R3, #0
		MOV	R4, #3
		SWI	OS_ReadVarVal+XBit	; Variable lesen
		MOVVS	R2, #0			; Default: 0 Bytes langer Pfad

mfn_cp_name	ADR	R4, mfn_filename
mfn_loop	LDRBT	R0, [R4], #1		; Filenamen an Pfad haengen
		STRB	R0, [R1, R2]
		ADD	R2, R2, #1
		CMP	R0, #0
		BNE	mfn_loop

		LDMIA	SP, {R0,R4}		; Drive- und Partitionnummer
		ORR	R0, R4, R0, LSL #8	; Hex: DDPP, DD=Drive, PP=Part.
		ADD	R1, R1, R2		; Ende des Filenamens
		SUB	R1, R1, #1
		MOV	R2, #5			; max. Laenge
		SWI	OS_ConvertHex4+XBit

;		ADR	R0, mfn_file_path
;		SPrint	"Filename: "
;		SWI	OS_Write0+XBit
;		NLPrint

		ADR	R0, mfn_file_path
		LDMIA	SP!, {R1-R4, PC}^

mfn_flop_name	=	"Minix$Floppy", 0
mfn_hard_name	=	"Minix$Harddisk", 0
mfn_swap_name	=	"Minix$Swap", 0
mfn_filename	=	"drive", 0

mfn_file_path	%	VarSize


;==============================================================================
; swi_diskparms
;
; Holt Parameter der einzelnen Laufwerke
;
; R0=-1 		==> R0=Pointer auf Liste mit Laufwerksnummern,
;			    Bit  0-15: Partition-# (im Moment 00-FF)
;			    Bit 16-31: Drive-#     (im Moment 0-2)
;			    mit -1 abgeschlossen
; R0=Laufwerksnummer
; R1=Partitionnummer
; R2=Pointer auf Infoblock, der gefuellt werden soll
;
; in  : R0, R1, R2
; out : R0=Pointer auf Information
; uses: -
;==============================================================================
swi_diskparms
		STMDB	SP!, {R1-R6, LK}
		CMP	R0, #0
		BGE	dpar_drives		; Info ueber ein Drive!
		ADR	R0, dpar_drvlist
		MOV	R1, #Flop_Dev		; Floppy
		MOV	R2, #NumFdDrives	; Anzahl zu testender Part.
		BL	dpar_tst_drv
		MOV	R1, #Hard_Dev		; HD
		MOV	R2, #NumHdDrives
		BL	dpar_tst_drv
		MOV	R1, #Swap_Dev		; Swap
		MOV	R2, #NumSpDrives
		BL	dpar_tst_drv
		MOV	R1, #-1
		STR	R1, [R0]
		ADR	R0, dpar_drvlist
		B	dpar_exit

dpar_drives	MOV	R6, R2			; Pointer auf Info-Struktur
		MOV	R2, R1			; Partition-#
		MOV	R1, R0			; LW-#
		BL	make_filename
		MOV	R1, R0
		MOV	R0, #&11		; File Info lesen
		SWI	OS_File+XBit
		CMP	R0, #1
		BNE	dpar_err		; File existiert nicht!
		STR	R4, [R6, #DP_DSize]	; Disk-Groesse
		MOV	R1, #512
		STR	R1, [R6, #DP_SSize]	; Sektor-Groesse

dpar_exit	LDMIA	SP!, {R1-R6, LK}
		BICS	PC, LK, #OverflowFlag

dpar_err	LDMIA	SP!, {R1-R6, LK}
		ORRS	PC, LK, #OverflowFlag

dpar_drvlist	%	(NumFdDrives+NumHdDrives+NumSpDrives+1)*4


;==============================================================================
; dpar_tst_drv
;
; Testet, welche Partitionen von R1 existieren und traegt deren
; Nummern in die Liste ein, auf die R0 zeigt. R0 zeigt am Ende auf die
; naechste freie Position in der Liste.
;
; in  : R0	Pointer auf freie Position in der Liste
;	R1	Nummer des zu testenden Drives
;	R2	Anzahl zu testender Partitionen
; out : R0	Pointer auf Listende
; uses: R0-R4
;==============================================================================
dpar_tst_drv
		STMDB	SP!, {LK}
		SUB	R3, R2, #1		; max. Partition-#
		MOV	R2, #0			; Start bei Partition 0
		B	dpar_tst_nxtdrv

dpar_tst_loop1	STMDB	SP!, {R0}
		BL	make_filename
		BL	exist_file
		LDMIA	SP!, {R0}
		ORR	R4, R2, R1, LSL #16
		STRVCT	R4, [R0], #4		; Part-/LW-nummer abspeichern
		ADD	R2, R2, #1		; naechste Partition
dpar_tst_nxtdrv	CMP	R2, R3
		BLE	dpar_tst_loop1

dpar_tst_exit	LDMIA	SP!, {PC}^


;==============================================================================
; exist_file
;
; Testet, ob die Datei, auf deren Namen R0 zeigt, existiert.
; VC bei Existenz, ansonsten VS
;
; in  : R0	Pointer auf Filenamen
; out : -
; uses: -
;==============================================================================
exist_file
		STMDB	SP!, {R0-R5, LK}
		MOV	R1, R0
		MOV	R0, #&11		; read file info
		SWI	OS_File+XBit
		CMP	R0, #0
		LDMIA	SP!, {R0-R5, LK}
		BICNES	PC, LK, #OverflowFlag	; return VC
		ORREQS	PC, LK, #OverflowFlag	; return VS


;==============================================================================
; swi_set_vector
;
; in  : R0		Vektornummer
;	R1		Adresse, auf die verzweigt werden soll
;                       (bzw. 0 falls Vektor freigegeben werden soll)
; out : -
; uses: -
;==============================================================================
swi_set_vector
		STMDB	SP!, {R0-R2,LK}
		CMP	R0, #SV_ClockVec
		BNE	sve_notclock
		CMP	R1, #0			; Vektor freigeben?
		BEQ	sve_relclk		; ja!
		STR	R1, sve_clockvec	; Zieladresse retten
		MOV	R0, #&1C		; TickerV = Clock Vektor
		ADR	R1, sve_clock
		MOV	R2, #0
		SWI	OS_Claim+XBit
		B	sve_end
sve_relclk
		MOV	R0, #&1C
		ADR	R1, sve_clock
		MOV	R2, #0
		SWI	OS_Release+XBit
		B	sve_end
sve_notclock
		CMP	R0, #SV_KeybdVec
		BNE	sve_notkeybd
		CMP	R1, #0			; Vektor freigeben?
		BEQ	sve_relkbd		; ja!
		STR	R1, sve_keybdvec
		MOV	R0, #&10		; EventV = Event Vektor
		ADR	R1, sve_keybd
		MOV	R2, #0
		SWI	OS_Claim+XBit		; Eventvektor belegen
		MOV	R0, #&0E		; Enable Event
		MOV	R1, #11			; Key up/down event
		SWI	OS_Byte+XBit		; Keyboardevent freigeben
		B	sve_end
sve_relkbd
		MOV	R0, #&0D		; Disable Event
		MOV	R1, #11			; Key up/down event
		SWI	OS_Byte+XBit		; Event sperren
		MOV	R0, #&1C
		ADR	R1, sve_keybd
		MOV	R2, #0
		SWI	OS_Release+XBit		; Vektor freigeben
		B	sve_end
sve_notkeybd
sve_end
		LDMIA	SP!, {R0-R2,LK}
		BICS	PC, LK, #OverflowFlag

; Interruptroutine: wird von Arthur aufgerufen und sichert Parameter und
; Startadresse der Interruptroutine im irq-Buffer.
; R12 ist unbelegt und wird vom Arthur gesichert
sve_clock
		LDR	LK, ioe_queue_ptr
		ADRL	R12, ioe_queue
		SUB	R12, LK, R12
		CMP	R12, #IRQQSize*4*4
		BGE	sve_clock_end
		LDR	R12, sve_clockvec
		STMIA	LK!, {R0-R2,R12}
		STR	LK, ioe_queue_ptr
sve_clock_end	LDMIA	R13!, {PC}

; Interruptroutine: leitet Tastaturinterrupt ins Minix-OS weiter
sve_keybd
		LDR	LK, ioe_queue_ptr
		ADRL	R12, ioe_queue
		SUB	R12, LK, R12
		CMP	R12, #IRQQSize*4*4
		BGE	sve_keybd_end
		LDR	R12, sve_keybdvec
		MOV	R0, R2			; R0=code, R1=up/down
		STMIA	LK!, {R0-R2,R12}
		STR	LK, ioe_queue_ptr
sve_keybd_end	LDMIA	R13!, {PC}

sve_keybdvec	DCD	0
sve_clockvec	DCD	0			; Zieladresse


;==============================================================================
; swi_special
;
; in  :	R0	0	Setzt Keyboard-Leds
;			R1: Bit0=Caps, Bit1=Num, Bit2=Scroll
;			R2: Bitmaske: 1: Wert setzen, 0: Wert nicht veraendern
;	R0	1	Read Timer, Ergebnis in R0
;	R0	2	Reset erzeugt Hard-Reset
; out : R0	evtl. Rueckgabewert
; uses: -
;==============================================================================
swi_special
		STMDB	SP!, {R1-R3,LK}
		CMP	R0, #SP_SetLeds
		BNE	spc_timer
		MOV	R0, #0
		TST	R1, #1		; Caps-Lock?
		ORRNE	R0, R0, #&10	; Arthur schaltet nochmals um!
		TST	R1, #2		; Num-Lock?
		ORRNE	R0, R0, #&04	; Arthur schaltet nochmals um!
		TST	R1, #4		; Scroll-Lock?
		ORRNE	R0, R0, #&02
		MOV	R1, R0
		MOV	R0, #&FF
		TST	R2, #1
		BICNE	R0, R0, #&10
		TST	R2, #2
		BICNE	R0, R0, #&04
		TST	R2, #4
		BICNE	R0, R0, #&02
		BIC	R1, R1, R0
		MOV	R2, R0
		MOV	R0, #&CA	; Read/Write keyboard status byte
		SWI	OS_Byte+XBit	; Status nun setzen
		MOV	R0, #&76	; Reflect keyboard status in LED's
		SWI	OS_Byte+XBit
		B	spc_end
spc_timer
		CMP	R0, #SP_ReadTimer
		BNE	spc_reset
		MOV	R1, #IOC_BASE
		STRB	R1, [R1, #IOC_T1LATCH]
		LDRB	R0, [R1, #IOC_T1LATCHL]
		LDRB	R1, [R1, #IOC_T1LATCHH]
		ORR	R0, R0, R1, LSL #8
		B	spc_end
spc_reset
		CMP	R0, #SP_Reset
		BNE	spc_realtime
		MOV	R0, #ROMStart
		LDR	R0, [R0]		; Holt Branch-Befehl auf Reset
		BIC	R0, R0, #&FF000000
		ADD	R0, R0, #2		; Prefetch-Korrektur
		MOV	R0, R0, LSL #2
		ORRS	PC, R0, #&0C000003	; Interrupts sperren & SVC-Mode
		B	spc_end

spc_realtime
		CMP	R0, #SP_RealTime
		BNE	spc_end
		STMDB	SP!, {R4-R7}
		MOV	R0, #14			; Read Realtime seit 1900
		ADR	R1, rt_buf
		MOV	R2, #3			; reason code
		STR	R2, [R1, #0]
		SWI	OS_Word
		LDR	R2, [R1, #0]
		MOV	R0, #0
		LDRB	R0, [R1, #4]
		MOV	R1, R2
		MOV	R2, #0			; durch 100 teilen
		MOV	R3, #100
		BL	div_64
		LDR	R1, =&83AA7E80		; #Sekunden 1900 - 1970
		SUB	R0, R5, R1
		LDMIA	SP!, {R4-R7}
		B	spc_end

spc_end		LDMIA	SP!, {R1-R3,LK}
		BICS	PC, LK, #OverflowFlag

rt_buf		%	8


div_64		; dividiert R01 / R23 = R45 Rest R01
		MOV	R4, #0			; Ergebnis
		MOV	R5, #0
		MOV	R6, #0			; R67 = Stellenwertigkeit
		MOV	R7, #1
div_64_1	ADDS	R7, R7, R7		; R67 << 1
		ADC	R6, R6, R6
		ADDS	R3, R3, R3		; R23 << 1
		ADC	R2, R2, R2
		TST	R2, #&80000000
		BEQ	div_64_1
div_64_2	SUBS	R1, R1, R3		; R01 -= R23
		SBCS	R0, R0, R2
		BCS	div_64_3
		ADDS	R1, R1, R3		; R01 += R23
		ADC	R0, R0, R2
		B	div_64_4
div_64_3	ADDS	R5, R5, R7		; R45 += R
		ADC	R4, R4, R6
div_64_4	MOVS	R2, R2, LSR #1		; R23 >> 1
		MOVS	R3, R3, RRX
		MOVS	R6, R6, LSR #1		; R67 >> 1
		MOVS	R7, R7, RRX
		TEQ	R6, #0			; R67 == 0?
		TEQEQ	R7, #0
		BNE	div_64_2
		MOVS	PC, LK


;==============================================================================
; swi_get_para
;
; Routine holt Systemparameter lt. Beschreibung fuer Routine get_para
;
; in  :	R0		Pointer auf Datenblock, der gefuellt werden soll
; out :	-
; uses: -
;==============================================================================
swi_get_para
		STMDB	SP!, {LK}
		BL	get_para
		LDMIA	SP!, {LK}
		BICS	PC, LK, #OverflowFlag


;==============================================================================
; swi_get_map
;
; Die Routine stellt die belegten logischen Pagenummern fest und sucht
; die jeweils dazu gehoerige physikalische Pagenummer.
; Ausgabe der Ergebnisse in ein Array, auf das R0 zeigt,
; muss Platz fuer alle Adressen der physikalischen Pages haben
;
; in  :	R0		Pointer auf das Array von Worten f.d. Pageadressen
; out : -
; uses: -
;==============================================================================
swi_get_map
		STMDB	SP!, {R1-R5, LK}

		TEQP	PC, #&0C000003		; Disable (F)IRQ

		MOV	R1, #VecDataAbort
		ADR	R2, my_data_abort
		LDR	R5, [R1, #0]
		GenBra	R1, R2
		STR	R2, [R1, #0]

		LDR	R1, act_ppg_num		; #physik. Pages
		MOV	R2, #-1
sgm_loop01	SUBS	R1, R1, #1
		STR	R2, [R0, R1, LSL #2]	; Feld mit -1 fuellen
		BGT	sgm_loop01

		MOV	R1, #0			; Pagenummer
		LDR	R2, act_pg_size		; Pagesize
		LDR	R3, log_base		; log. Pageadresse
sgm_loop02	MOV	R4, #0
		STR	R4, abort_flag

		LDR	R4, [R3]		; -> evtl DATA-Abort-Exception
		LDR	R4, abort_flag
		CMP	R4, #0
		BNE	sgm_next_page		; ungueltige Page!!!

		ADD	R1, R1, #1		; Eine weitere Page gefunden
		MOV	R4, R0			; Arrayadresse retten
		MOV	R0, R3
		BL	find_page

		CMP	R0, #-1			; Page gefunden?
		STRNE	R3, [R4, R0, LSL #2]
		MOV	R0, R4
sgm_next_page	ADD	R3, R3, R2		; naechste logische Page

		LDR	R4, log_end
		CMP	R3, R4
		BLT	sgm_loop02

		MOV	R1, #VecDataAbort	; alten Vektor restaurieren
		STR	R5, [R1, #0]

		LDMIA	SP!, {R1-R5, LK}
		BICS	PC, LK, #OverflowFlag


;==============================================================================
; swi_free_pages
;
; liefert eine Liste der freien Pagenummern, die das AOS dem Benutzer
; als freien Speicher ueberlaesst. R0 ist ein Zeiger auf einen Speicher-
; bereich, in den die Nummern geschrieben werden koennen.
; Das Ende der Liste wird mit -1 angezeigt.
;
; in  :	R0		Pointer auf Puffer
; out :	-
; uses: 0.5K auf dem Stack
;==============================================================================
swi_free_pages
		STMDB	SP!, {R0-R4, LK}

		MOV	R1, SP			; alten SP retten
		LDR	R3, act_ppg_num
		SUB	SP, SP, R3, LSL #2	; Workspace schaffen
		MOV	R0, SP
		BL	swi_get_map		; Pagemapping holen

		MOV	R2, #-1			; phys. Pagenummer
		LDR	R0, [R1]		; Pointer auf Listenbereich

		B	frp_loop02
frp_loop01	LDR	R4, [SP, R2, LSL #2]	; log. Pageadr. aus Tab. holen
		CMP	R4, #ApplWorkspace	; Systembereich?
		BLT	frp_loop02		; ja ->
		CMP	R4, #RAMDisc		; ApplWorkspace?
		STRLTT	R2, [R0], #4		; ja -> speichern!
frp_loop02	ADD	R2, R2, #1
		CMP	R2, R3
		BLT	frp_loop01
		MOV	R2, #-1
		STR	R2, [R0]		; Listenende!

		ADD	SP, SP, R3, LSL #2

		LDMIA	SP!, {R0-R4, LK}
		BICS	PC, LK, #OverflowFlag


;==============================================================================
; swi_set_trnsl
;
; Setzt einen Eintrag in der Translation-Tabelle
; Die Werte duerfen die entspr. Wertebereiche nicht ueberschreiten!
; in  : R0		physikalische Pagenummer (0-127)
;	R1		logische Pagenummer (0-8192/4096/2048/1024)
;	R2		Page-Protection-Code (0-3)
; out :
; uses: -
;==============================================================================
swi_set_trnsl
		STMDB	SP!, {R0-R3}
		LDR	R3, cod_pg_size
		ADD	PC, PC, R3, LSL #2
str_trnl_tab	DCD	0
		B	str_trnl_type0
		B	str_trnl_type1
		B	str_trnl_type2
		B	str_trnl_type3

str_trnl_type0	AND	R3, R0, #&7F		; phys. Pagenummer
		ORR	R3, R3, R2, LSL #8	; PPL
		LDR	R2, =&1FFF
		AND	R1, R1, R2
		ORR	R3, R3, R1, LSL #10
		ORR	R3, R3, #AdrTrnslBase
		STRB	R3, [R3, #0]
		LDMIA	SP!, {R0-R3}
		BICS	PC, LK, #OverflowFlag

str_trnl_type1	AND	R3, R0, #&3F
		MOV	R3, R3, LSL #1		; Bit 0-5 nach Bit 1-6
		TST	R0, #&40		; Bit 6 nach Bit 0
		ORRNE	R3, R3, #1
		ORR	R3, R3, R2, LSL #8	; PPL
		AND	R2, R1, #&C00		; Bit 10 und 11
		ORR	R3, R3, R2		; 1:1 uebertragen
		LDR	R2, =&3FF
		AND	R1, R1, R2
		ORR	R3, R3, R1, LSL #13	; Bit 0-9 nach Bit 13-22
		ORR	R3, R3, #AdrTrnslBase
;		SPrint	"Setting in MEMC: "
;		HPrint	R3
;		NLPrint
		STRB	R3, [R3, #0]
		LDMIA	SP!, {R0-R3}
		BICS	PC, LK, #OverflowFlag

str_trnl_type2	AND	R3, R0, #&1F
		MOV	R3, R3, LSL #2		; Bit 0-4 nach Bit 2-6
		TST	R0, #&20		; Bit 5 nach Bit 0
		ORRNE	R3, R3, #1
		TST	R0, #&40		; Bit 6 nach Bit 1
		ORRNE	R3, R3, #2
		ORR	R3, R3, R2, LSL #8	; PPL
		AND	R2, R1, #&600
		ORR	R3, R3, R2, LSL #1	; Bit 9-10 nach Bit 10-11
		LDR	R2, =&1FF
		AND	R1, R1, R2
		ORR	R3, R3, R1, LSL #14	; Bit 0-8 nach Bit 14-22
		ORR	R3, R3, #AdrTrnslBase
		STRB	R3, [R3, #0]
		LDMIA	SP!, {R0-R3}
		BICS	PC, LK, #OverflowFlag

str_trnl_type3	AND	R3, R0, #&0F
		MOV	R3, R3, LSL #3		; Bit 0-3 nach Bit 3-6
		TST	R0, #&10		; Bit 4 nach Bit 0
		ORRNE	R3, R3, #1
		TST	R0, #&20		; Bit 5 nach Bit 2
		ORRNE	R3, R3, #4
		TST	R0, #&40		; Bit 6 nach Bit 1
		ORRNE	R3, R3, #2
		ORR	R3, R3, R2, LSL #8	; PPL
		AND	R2, R1, #&300
		ORR	R3, R3, R2, LSL #2	; Bit 8-9 nach Bit 10-11
		AND	R1, R1, #&FF
		ORR	R3, R3, R1, LSL #15	; Bit 0-7 nach Bit 15-22
		ORR	R3, R3, #AdrTrnslBase
		STRB	R3, [R3, #0]
		LDMIA	SP!, {R0-R3}
		BICS	PC, LK, #OverflowFlag


;==============================================================================
; swi_os_mode
;
; Schaltet den OS-Mode des MEMC ein (R0 <> 0) oder aus (R0 = 0)
; in  : R0	= 0: OS-Mode aus, ansonsten ein
; out : R0	neuer Wert des MEMC Control Registers
; uses: -
;==============================================================================
swi_os_mode
		TEQ	R0, #0
		LDR	R0, memc_control
		ORRNE	R0, R0, #OS_Mode
		BICEQ	R0, R0, #OS_Mode
		STR	R0, memc_control
		STRB	R0, [R0, #0]
		BICS	PC, LK, #OverflowFlag


;==============================================================================
; swi_set_map
;
; Setzt Pagemapping neu. R0 ist ein Zeiger auf eine Liste von Worten, die mit
;	-1 abgeschlossen ist. In den oberen 16 Bits steht die Nummer der
;	physikalischen Page, die auf die logische Page gemappt wird, die in
;	den unteren 16 Bits steht.
; in  : R0	Pointer auf die Liste
;	R1	Page Protection Level
; out : -
; uses: -
;==============================================================================
swi_set_map
		STMDB	SP!, {R0-R3,LK}
		MOV	R2, R1
		MOV	R3, R0
		B	smp_test
smp_loop	BIC	R1, R0, #&FF000000
		BIC	R1, R1, #&00FF0000
		MOV	R0, R0, LSR #16
;		HPrint	R0
;		SPrint	" --> "
;		HPrint	R1
;		SPrint	", prot = "
;		HPrint	R2
;		NLPrint
		BL	swi_set_trnsl
smp_test	LDRT	R0, [R3], #4
		CMP	R0, #-1
		BNE	smp_loop
		LDMIA	SP!, {R0-R3,LK}
		BICS	PC, LK, #OverflowFlag


;==============================================================================
; get_para
;
; Holt rechnerspezifische Parameter und setzt diese in dem Datenblock,
; auf den R0 zeigt:
; R0 ->	+&00	min_pg_size	minimale Pagegroesse in Bytes
;	+&04	max_pg_size	maximale Pagegroesse in Bytes
;	+&08	act_pg_size	aktuelle Pagegroesse in Bytes
;	+&0C	cod_pg_size	log_2(act_pg_size/min_pg_size)
;	+&10	min_ppg_num	minimale Anzahl physikalischer Pages
;	+&14	max_ppg_num	maximale Anzahl physikalischer Pages
;	+&18	act_ppg_num	aktuelle Anzahl physikalischer Pages
;	+&1C	min_lpg_num	minimale Anzahl logischer Pages
;	+&20	max_lpg_num	maximale Anzahl logischer Pages
;	+&24	act_lpg_num	aktuelle Anzahl logischer Pages
;	+&28	phys_base	Basisadresse der physikalischen Pages
;	+&2C	phys_end	Endadresse der physikalischen Pages (ausschl.)
;	+&30	log_base	Basisadresse der logischen Pages
;	+&34	log_end		Endadresse der logischen Pages (ausschl.)
;	+&38	free_area	Startadresse freies RAM
;	+&3C	top_ram		Endadresse freies RAM
;==============================================================================
get_para
		STMDB	SP!, {R1-R3}
		LDR	R1, =MinPageSize
		STR	R1, [R0, #GP_MinPgSize]
		LDR	R1, =MaxPageSize
		STR	R1, [R0, #GP_MaxPgSize]
		MOV	R1, #0			; Pagesize holen
		MOV	R2, #0
		LDR	R1, memc_control
		AND	R1, R1, #&C
		MOV	R1, R1, LSR #2
		STR	R1, [R0, #GP_CodPgSize]
		MOV	R2, #4096
		MOV	R1, R2, LSL R1
		STR	R1, [R0, #GP_ActPgSize]
		LDR	R1, =MaxPages
		STR	R1, [R0, #GP_MinPPgNum]
		STR	R1, [R0, #GP_MaxPPgNum]
		STR	R1, [R0, #GP_ActPPgNum]
		LDR	R1, =1024
		STR	R1, [R0, #GP_MinLPgNum]
		LDR	R1, =8192
		STR	R1, [R0, #GP_MaxLPgNum]
		LDR	R2, [R0, #GP_CodPgSize]
		MOV	R1, R1, LSR R2
		STR	R1, [R0, #GP_ActLPgNum]
		LDR	R1, =PhysPageBase
		STR	R1, [R0, #GP_PhysBase]
		STR	R1, [R0, #GP_LogEnd]
		LDR	R2, [R0, #GP_ActPPgNum]
		LDR	R3, [R0, #GP_ActPgSize]
		MLA	R1, R2, R3, R1
		STR	R1, [R0, #GP_PhysEnd]
		LDR	R1, =LogPageBase
		STR	R1, [R0, #GP_LogBase]
		LDR	R1, =&8000
		STR	R1, [R0, #GP_FreeArea]
		LDR	R1, =&1000000
		STR	R1, [R0, #GP_TopRam]
		LDMIA	SP!, {R1-R3}
		MOVS	PC, LK


;==============================================================================
; my_data_abort
;
; neue, temporaere DATA-ABORT-Routine
; Setzt nur ein Flag (abort_flag) auf 1, um anzuzeigen, dass
; sie aufgerufen wurde.
;
; in  : -
; out : abort_flag
;==============================================================================
my_data_abort
		STMDB	SP!, {R0}
		MOV	R0, #1
		STR	R0, abort_flag		; Abort-Flag setzen
		LDMIA	SP!, {R0}
		SUBS	PC, LK, #4		; Falschen Befehl uebergehen


;==============================================================================
; find_page
;
; Sucht im physikalischen Adressbereich nach einer Page,
; auf die R0 zeigt (R0 = Zeiger auf Page im logischen Adressbereich)
;
; in  : R0		Zeiger auf logische und gueltige Pageadresse
;	act_pg_size	Pagegroesse in Bytes
;			Aufruf im SVC-Modus!
; out : R0		Nummer der gefundenen physikalischen Page
;			(-1 falls nicht gefunden)
;==============================================================================
find_page
		STMDB	SP!, {R1-R7}

		TEQP	PC, #&0C000003		; Disable Interrupts

		LDR	R2, act_pg_size		; Suchraster
		LDR	R3, act_ppg_num		; Wieviel Suchschritte
		ADD	R4, R0, R2		; Ende der log. Page berechnen
		LDR	R5, [R4, #-4]		; Letztes Wort retten
		LDR	R6, =PageMagic
		STR	R6, [R4, #-4]		; Magic-Nr am Pageende setzen

		LDR	R7, phys_base		; Ende der 1. physik. Page
		ADD	R7, R7, R2		; berechnen
		SUB	R7, R7, #4

fp_loop1	LDR	R1, [R7]
		CMP	R1, R6			; Magic-Nr gefunden?
		BEQ	fp_magic_found		; ja ->
		ADD	R7, R7, R2		; naechste Page
		SUBS	R3, R3, #1
		BGT	fp_loop1

fp_page_not_fnd	MOV	R0, #-1
		STR	R5, [R4, #-4]
		B	fp_ret

fp_magic_found	STR	R5, [R4, #-4]		; letztes Wort restaurieren
		LDR	R4, act_ppg_num		; gefundene Pagenummer berechnen
		SUB	R0, R4, R3		; [0..MAX_PAGES-1]

fp_ret		LDMIA	SP!, {R1-R7}
		MOVS	PC, LK


		LTORG

;
;============== DATA SECTION ==================================================
;

		ALIGN

		^	0, R10		; R10=Pointer auf aktuellen WS, alles
					; Restliche ist relativ dazu adressiert

; System-Parameterblock, wird gefuellt durch get_para
data_start	EQU	@
sys_para_block	EQU	@
min_pg_size	#	4
max_pg_size	#	4
act_pg_size	#	4
cod_pg_size	#	4
min_ppg_num	#	4
max_ppg_num	#	4
act_ppg_num	#	4
min_lpg_num	#	4
max_lpg_num	#	4
act_lpg_num	#	4
phys_base	#	4
phys_end	#	4
log_base	#	4
log_end		#	4
free_area	#	4
top_ram		#	4
sys_para_end	EQU	@
		ASSERT	sys_para_end = sys_para_block + GP_Size


memc_control	#	4		; Kopie des MEMC-Contrl-Reg's
swi_ret_mode	#	4		; Flag fuer SVC-Umschaltung
abort_flag	#	4
		[	DEBUG > 0
macbuf		#	4
macbuffer	#	MACBUFSIZE__
		]

map		#	4*MaxPages
stackbeg	#	StackSize
stackend	EQU	@
data_end	EQU	@
data_size	EQU	data_end - data_start

end		EQU	.

		END
