; This file contains a number of assembly code utility routines needed by the
; kernel.  They are:


		GET	$.arm.aslib.register
		GET	$.arm.aslib.constants
		GET	$.arm.aslib.macros
		GET	^.^.arm.h.aosdriver
		GET	^.^.arm.h.asconst


		AREA	|C$$code|, CODE, READONLY

		EXPORT	klib_2hook
		EXPORT	set_vec
		EXPORT	test_and_set
		EXPORT	reset
		EXPORT	wait_retrace
		EXPORT	phys_copy

		EXPORT	out_byte
		EXPORT	in_byte



;*===========================================================================*
;*				klib_2hook				     *
;*===========================================================================*
; PUBLIC void klib_prot_mode_init();
; Initialize klib protected real mode for real or protected mode,
; Do nothing for real mode.
; For protected mode, load idt, task reg and flags.
; For CHIP==VL86C010 it's a dummy routine

klib_2hook
		MOVS	PC, LK


;*===========================================================================*
;*				set_vec					     *
;*===========================================================================*
; PUBLIC void set_vec(int vec_no, void * vec_addr);

set_vec
		SWI	MX_SetVector
		MOVS	PC, LK


;*===========================================================================*
;*				test_and_set				     *
;*===========================================================================*
; PUBLIC int test_and_set(int *flag);
; Set the flag to TRUE, indivisibly with getting its old value.
; Return old flag.

test_and_set
		MOV	A4, LK
		SWI	MX_EnterSVC
		TEQP	PC, #FIRQFlag:OR:IRQFlag:OR:SVCMode
		LDR	A2, [A1, #0]
		MOV	A3, #1
		STR	A3, [A1, #0]
		MOV	A1, A2
		MOVS	PC, A4


;*===========================================================================*
;*				reset					     *
;*===========================================================================*
; PUBLIC void reset();
; Reset the system.

reset
		MOV	R0, #SP_Reset
		SWI	MX_Special
reset_1		B	reset_1


;*===========================================================================*
;*			      wait_retrace				     *
;*===========================================================================*
; PUBLIC void wait_retrace();
; Wait for the *start* of retrace period.
; We can't afford to worry about interrupts.
; On the ARM there is nothing to do.

wait_retrace
		MOVS	PC, LK


;*===========================================================================*
;*				phys_copy				     *
;*===========================================================================*
; PUBLIC void phys_copy(phys_bytes source, phys_bytes destination,
;			phys_bytes bytecount);
; Copy a block of physical memory.

phys_copy
		CMP	A3, #32		; more bytes to copy?
		BLT	pc_short_test	; no

		TST	A1, #3		; source: word aligned address?
		TSTEQ	A2, #3		; destination: word aligned address?
		BNE	pc_short_test

		STMDB	SP!, {V1-V4}
		B	pc_long_test
pc_long_loop	LDMIA	A1!, {V1-V4}	; load 4 words
		STMIA	A2!, {V1-V4}	; store them
pc_long_test	SUBS	A3, A3, #16	; another 16 Bytes to copy?
		BGE	pc_long_loop	; yes!
		LDMIA	SP!, {V1-V4}
		ADDS	A3, A3, #16
		MOVEQS	PC, LK		; no more single bytes to copy
		B	pc_short_test

pc_short_loop	LDRB	A4, [A1, A3]
		STRB	A4, [A2, A3]
pc_short_test	SUBS	A3, A3, #1	; another byte to copy?
		BGE	pc_short_loop
		MOVS	PC, LK



;*===========================================================================*
;*				in_byte					     *
;*===========================================================================*
; PUBLIC unsigned in_byte(port_t port);
; Read an (unsigned) byte from the i/o port  port  and return it.
;
in_byte
		MOVS	PC, LK


;*===========================================================================*
;*				out_byte				     *
;*===========================================================================*
; PUBLIC void out_byte(port_t port, int value);
; Write  value  (cast to a byte)  to the I/O port  port.
;
out_byte
		MOVS	PC, LK




;*===========================================================================*
;*				data					     *
;*===========================================================================*

		AREA	|C$$data|, DATA

splimit		DCD	0


		END
