; This boot program loads the five pieces of Minix into the memory,
; relocates them and clears the bss segment.
; It then starts fsck.
; In detail:
; First, if needed, the program loads the Minix-AOS Driver (minixdrv-
; module) into memory and starts it.
; Then it fetches address and size of the free memory and calculates
; the total size of the five parts of Minix (kernel, mm, fs, init, fsck).
; If they will fit, the program copies the loader (a part of itself)
; to the top of the free memory (highmem) and starts it.
; The loader then loads the five parts. Each part will be relocated
; and then the bss segment cleared. Then the loader patches the
; addresses of the other parts into the first few bytes of the bss
; segment of each part:
; kernel_text	DCD	0
; kernel_data	DCD	0
; kernel_bss	DCD	0
; kernel_end	DCD	0
; mm_text	DCD	0
; mm_data	DCD	0
; mm_bss	DCD	0
; mm_end	DCD	0
; fs_text	DCD	0
; fs_data	DCD	0
; fs_bss	DCD	0
; fs_end	DCD	0
; init_text	DCD	0
; init_data	DCD	0
; init_bss	DCD	0
; init_end	DCD	0
; fsck_text	DCD	0
; fsck_data	DCD	0
; fsck_bss	DCD	0
; fsck_end	DCD	0



		GET	$.arm.aslib.register
		GET	$.arm.aslib.constants
		GET	$.arm.aslib.arthur
		GET	$.minix.usr.arm.h.aosdriver
		GET	$.arm.aslib.macros


; File-Header: Offsets into the executable file where to find the
;              different parts of the program
FILE_RELOC	EQU	&4
FILE_CLR_BSS	EQU	&8
FILE_MAIN	EQU	&C
FILE_TEXT_LEN	EQU	&14
FILE_DATA_LEN	EQU	&18
FILE_BSS_LEN	EQU	&20
FILE_RELOC_BASE	EQU	&28


		ORG	&8000



;==============================================================================
; main entry of the minix loader
;==============================================================================
boot_minix
		InitMem
		InitSP	#&400			; 1K stacksize
		InitMac

; Set AOS options (keyboard handler etc...)
		MOV	R0, #4			; set cursor key status
		MOV	R1, #1			; cursor keys as ASCII code
		SWI	OS_Byte+XBit
		MOV	R0, #&C8		; set break/escape effect
		MOV	R1, #1			; disable escape
		MOV	R2, #0
 		SWI	OS_Byte+XBit
		MOV	R0, #&CA		; set keyboard state
		MOV	R1, #&10
		MOV	R2, #0
		SWI	OS_Byte+XBit
		MOV	R0, #&76		; reflect status in LEDs
		SWI	OS_Byte+XBit
		MOV	R0, #&F7		; set break keys
		MOV	R1, #&8A
		MOV	R2, #0
		SWI	OS_Byte+XBit

		MOV	R0, #0			; close files
		MOV	R1, #0			; close them all!
		SWI	OS_Find+XBit

; Load AOS driver if not present
		ADRL	R0, greet
		SWI	OS_Write0+XBit
		ADRL	R0, module_name		; search module
		ADRL	R1, path_var_name
		BL	search_path
		CMP	R0, #0
		BEQ	driv_not_fnd
		MOV	R1, R0			; pointer to filename
		MOV	R0, #0			; run module
		SWI	OS_Module+XBit		; load & run minix driver
		BVC	driver_present
		SPrint	"error: can't load Minix Driver "
		SPrint	"(RMA too small?)"
		NLPrint
		SWI	OS_Exit

driv_not_fnd	SPrint	"error: can't find Minix Driver: "
		ADRL	R0, module_name
		SWI	OS_Write0+XBit
		NLPrint
		SWI	OS_Exit

; AOS driver is now present:
; get the system parameters and search for the biggest part of the memory
driver_present
		MAlloc	R0, #GP_Size		; buffer for system parameters
		BVS	memory_ovfl
		STR	R0, sys_para
		SWI	MX_GetParameters+XBit	; get them
		BVS	driver_error

		LDR	R0, sys_para
		LDR	R0, [R0, #GP_ActPPgNum]
		MOV	R0, R0, LSL #2		; *4 bytes for pagemap table
		MAlloc	R0, R0
		BVS	memory_ovfl
		STR	R0, map

		SWI	MX_GetMemoryMap+XBit	; fetch page map
		BVS	driver_error

		LDR	R0, sys_para
		LDR	R1, [R0, #GP_ActPgSize]
		LDR	R2, [R0, #GP_FreeArea]
		LDR	R6, [R0, #GP_ActPPgNum]
		LDR	R0, map
		MOV	R3, #0			; number of pages
search_next_pg	MOV	R4, #0			; array index
search_page	LDR	R5, [R0, R4, LSL #2]
		ADD	R4, R4, #1
		CMP	R4, R6
		BGE	end_of_list		; end of list -> page not fnd!
		CMP	R5, R2
		BNE	search_page		; didn't find yet

		ADD	R3, R3, #1		; found one more page
		ADD	R2, R2, R1		; search for next page
		B	search_next_pg

end_of_list	MUL	R4, R3, R1		; ==> free mem in bytes
		RPrint	"Bytes free: ", R4


; Get the sizes of the kernel, mm, fs, init and fsck
		LDR	R0, sys_para
		LDR	R5, [R0, #GP_FreeArea]	; start of free mem
		MOV	R6, R5			; keep in mind...

		ADRL	R0, kernel_name
		BL	get_psize
		BVS	file_error
		SUB	R8, R3, #1		; Clicksize-1 for alignment
		GetLog2	R3, R7


		MACRO
		AlStr	$dest			; align R5 and store to $dest
		ADD	R5, R5, R8		; align to next clicksize
		MOV	R5, R5, LSR R7
		MOV	R5, R5, LSL R7
		STR	R5, $dest
		MEND

		MACRO
		LoadMod	$mod, $text
		SPrint	"Load address for $text              : "
		ADRL	R0, $mod._name
		BL	get_psize
		BVS	file_error
		AlStr	$mod._text
		SPrint	"t="
		HPrint	R5
		ADD	R5, R5, R0		; + textsize = start data
		AlStr	$mod._data
		SPrint	", d="
		HPrint	R5
		ADD	R5, R5, R1		; + datasize = start bss
		AlStr	$mod._bss
		SPrint	", b="
		HPrint	R5
		ADD	R5, R5, R2		; + bsssize = ende
		AlStr	$mod._end
		NLPrint
		MEND

		LoadMod	kernel, "kernel"
		LoadMod mm,     "mm    "
		LoadMod	fs,     "fs    "
		LoadMod	init,   "init  "
		LoadMod	fsck,   "fsck  "

; test for memory overflow
		SUB	R0, R5, R6		; amount of memory needed
		ADRL	R2, load_kernel
		ADRL	R3, load_kernel_end
		SUB	R3, R3, R2		; size of loader
		ADD	R0, R0, R3		; add size of loader
		CMP	R4, R0			; is it enough?
		BLE	memory_ovfl

;		RPrint	"Amount of memory needed: ", R0
;		RPrint	"Size of loader: ", R3

; copy loader to high mem and start it
		LDR	R0, sys_para
		LDR	R1, [R0, #GP_FreeArea]	; start of free memory
		ADD	R4, R1, R4		; end of free memory
;		RPrint	"Start of free memory: ", R1
;		RPrint	"End of free memory  : ", R4
		SUB	R4, R4, R3		; R4=start of loader
;		RPrint	"Start of loader     : ", R4
		MOV	R6, R4			; keep in mind

copy_loop	LDR	R5, [R2], #4		; copy the loader
		STR	R5, [R6], #4
		SUBS	R3, R3, #4
		BGT	copy_loop

		MOV	PC, R4			; start loader


;==============================================================================
; This subroutine fetches the sizes of the text- (R0), data- (R1) and
; bss-segments (R2). Returns also the click size of the kernel program in R3.
; A pointer to the name of the program must be given
; in R0. Returns VS on error, VC on success.
;==============================================================================
get_psize
		MOV	R1, R0			; pointer to filename
		MOV	R0, #&42		; open for readonly
		ADRL	R2, path_var_name	; default: Minix$Path
		SWI	OS_Find+XBit
		CMP	R0, #0
		ORREQS	PC, LK, #OverflowFlag	; return on error condition

		STMDB	SP!, {LK}
		MOV	R1, R0			; file handle
		MOV	R0, #FILE_TEXT_LEN
		BL	get_one_size
		BCS	get_psize_err
		STR	R0, get_psize_t
		MOV	R0, #FILE_DATA_LEN
		BL	get_one_size
		BCS	get_psize_err
		STR	R0, get_psize_d
		MOV	R0, #FILE_BSS_LEN
		BL	get_one_size
		BCS	get_psize_err
		STR	R0, get_psize_b
		MOV	R0, #FILE_MAIN		; get offset to main entry
		BL	get_one_size
		BCS	get_psize_err
		BIC	R0, R0, #&FF000000	; get offset from branch instr
		ADD	R0, R0, #2
		MOV	R0, R0, LSL #2
		ADD	R0, R0, #FILE_MAIN+4	; get file offset of CLICKSIZE
		BL	get_one_size
		BCS	get_psize_err
		STR	R0, get_click_size

		MOV	R0, #0			; close file
		SWI	OS_Find+XBit

		LDR	R3, get_click_size
		LDR	R2, get_psize_b
		LDR	R1, get_psize_d
		LDR	R0, get_psize_t
		LDMIA	SP!, {LK}
		BICS	PC, LK, #OverflowFlag

get_psize_err
		MOV	R0, #0			; close file on error
		SWI	OS_Find+XBit
		LDMIA	SP!, {LK}
		ORRS	PC, LK, #OverflowFlag

get_psize_t	DCD	0
get_psize_d	DCD	0
get_psize_b	DCD	0
get_click_size	DCD	0


;==============================================================================
; This subroutine reads a word (32 bit) from the file, given the file
; handle in R1. It reads the word beginning at offset R0.
; Returns the word in R0, CS on error, CC on success.
;==============================================================================
get_one_size
		STMDB	SP!, {R1-R4}
		MOV	R4, R0			; file offset
		MOV	R0, #3			; read bytes
		ADR	R2, gos_tmp
		MOV	R3, #4			; 4 bytes = 1 word
		SWI	OS_GBPB+XBit
		LDR	R0, gos_tmp
		LDMIA	SP!, {R1-R4}
		MOV	PC, LK			; CS on error condition
gos_tmp		DCD	0


;==============================================================================
; Searches a file (pointer to filename given in R0) along a specified
; path given by a variable (pointer to variablename given in R1).
; returns in R0 pointer to filename (=R0) or 0 on error
; or pointer to complete pathname with filename on success.
; In this case the buffer should be 'MFree'd if it isn't
; used any more.
;==============================================================================
search_path
		STMDB	SP!, {R0-R6}
		MOV	R0, R1			; pointer to variable name
		MAlloc	R1, #&100		; 256 bytes buffer
		CMP	R1, #0
		LDMEQIA	SP!, {R0-R6}
		MOVEQ	R0, #0
		MOVEQS	PC, LK			; return on error
		MOV	R2, #&100		; size of buffer
		MOV	R3, #0			; first call
		MOV	R4, #3			; expand string
		SWI	OS_ReadVarVal+XBit
		BVC	spth_var_found
spth_retfn	MFree	R1			; return only the filename
		LDMIA	SP!, {R0-R6}
		MOVS	PC, LK

spth_var_found
		MOV	R0, #0
		STRB	R0, [R1, R2]		; set end of string
		LDMIA	SP, {R0}		; fetch ptr to filename
		MAlloc	R2, #&100		; another buffer
		CMP	R2, #0
		BNE	spth_srch_path
		MFree	R1
		LDMIA	SP!, {R0-R6}
		MOV	R0, #0
		MOVS	PC, LK			; return on error

spth_srch_path	MOV	R3, R1			; source buffer (path list)
spth_srch_part	MOV	R4, R2			; destination buffer
		LDRB	R5, [R3], #1		; get next byte of the path
		CMP	R5, #0
		BNE	spth_srch
		MFree	R2			; file not found
		MFree	R1
		LDMIA	SP!, {R0-R6}
		MOV	R0, #0
		MOVS	PC, LK

spth_srch	CMP	R5, #0			; end of path list?
		SUBEQ	R3, R3, #1		; adjust for test above
		CMPNE	R5, #","		; or end of one part of the list
		BEQ	spth_test		; is there a file?
		STRB	R5, [R4], #1
		LDRB	R5, [R3], #1		; get next byte
		B	spth_srch
spth_test	MOV	R6, R0			; pointer to filename
spth_test_copy	LDRB	R5, [R6], #1
		STRB	R5, [R4], #1		; append filename
		CMP	R5, #0
		BNE	spth_test_copy
		STMDB	SP!, {R0-R3}
		MOV	R1, R2			; pointer to complete filename
		MOV	R0, #&43		; open for readonly/no path
		SWI	OS_Find+XBit
		CMP	R0, #0			; does the file exist?
		LDMEQIA	SP!, {R0-R3}
		BEQ	spth_srch_part		; test next part of path list
		MOV	R0, #0			; close file
		SWI	OS_Find+XBit
		LDMIA	SP!, {R0-R3}

		MFree	R1
		MOV	R0, R2
		ADD	SP, SP, #4		; 'pop R0'
		LDMIA	SP!, {R1-R6}		; pop the rest
		MOVS	PC, LK		



;==============================================================================
; Error routines
;==============================================================================
driver_error
		SPrint	"error: Minix Driver failure"
		NLPrint
		SWI	OS_Exit

memory_ovfl
		SPrint	"error: out of memory"
		NLPrint
		SWI	OS_Exit

file_error
		SPrint	"error: file not found"
		NLPrint
		SWI	OS_Exit


;==============================================================================
; data which are not copied to high mem
;==============================================================================

greet		=	13, "Booting MINIX 1.5.  Copyright 1988 "
		=	"Prentice-Hall, Inc.", 13, 10, 0
module_name	=	"MinixDrv", 0

		ALIGN

macbuf		DCD	0
sys_para	DCD	0
map		DCD	0		; max. # phys. Pages




;==============================================================================
;==============================================================================
; This is the loader.
; It will be copied to the end of the available memory and
; loads the five pieces of the kernel. It also patches the sizes
; of the single pieces to the first bytes of the data space of each
; part. At least it starts fsck.
;==============================================================================
;==============================================================================
		MACRO
		Load	$name
		ADRL	R0, $name._name
		STR	R0, act_name
		LDR	R1, $name._text
		LDR	R2, $name._data
		LDR	R3, $name._bss
		BL	load_file
		MEND

load_kernel
		ADRL	SP, loader_stack	; set inital stackpointer
		Load	kernel
		Load	mm
		Load	fs
		Load	init
		Load	fsck
		LDR	PC, fsck_text		; start fsck


;==============================================================================
; This routine loads the program pointed to by R0 into memory at address in R1.
; Then it relocates the program, moves the data segment to the address pointed
; to by R2 and clears the bss segment pointed to by R3 using own routines.
; To avoid double relocating and clearing the jumptable in the program
; will be modified accordingly.
;==============================================================================
load_file
		STMDB	SP!, {R0-R5, LK}
		MOV	R2, R1			; load destination address
		MOV	R1, R0			; pointer to filename
		MOV	R0, #&E
		MOV	R3, #0
		ADRL	R4, path_var_name
		SWI	OS_File+XBit		; load file
		LDMVSIA	SP!, {R0-R5, LK}	; fetch R0 on error
		BVS	kernel_error

; calculate the old start addresses of the data and bss segments relative
; to the start of the program in R10 and R11 and the differences to the
; new segments in R8 and R9
		LDMIA	SP, {R0-R5}
		LDR	R9, [R1, #FILE_RELOC_BASE]; old base
		LDR	R10, [R1, #FILE_TEXT_LEN] ; start of old data segment
		LDR	R11, [R1, #FILE_DATA_LEN]
		ADD	R11, R11, R10		; start of old bss segment
		ADD	R10, R10, R9
		ADD	R11, R11, R9

; start relocation
start_reloc	LDR	R0, [R1, #FILE_RELOC]
		MOV	R4, #&FB000000		; opcode for BLNV
		STR	R4, [R1, #FILE_RELOC]	; disable second relocation
		AND	R4, R0, #&FF000000	; get opcode
		CMP	R4, #&EB000000		; is it Branch with Link (BL)?
		LDMNEIA	SP!, {R0-R5, LK}	; fetch R0 on error
		BNE	kernel_error		; no! But this must be!
		AND	R0, R0, #&00FFFFFF	; get jump address
		ADD	R0, R1, R0, LSL #2
		ADD	R0, R0, #4		; end of reloc table contains
		LDR	R4, [R0, #4]		; # of symbols: get it
		CMP	R4, #0
		BEQ	move_dataseg		; nothing to do!
		SUB	R0, R0, R4, LSL #2	; get base-4 of reloc table
		STR	R1, [R1, #FILE_RELOC_BASE]

reloc_loop	LDR	R5, [R0, R4, LSL #2]	; get reloc address
		LDR	R6, [R1, R5]		; fetch adr to be reloct'd
		CMP	R6, R11			; pointer to bss segment?
		BLT	rel_loop1
		SUB	R6, R6, R11
		ADD	R6, R6, R3
		B	rel_loop3
rel_loop1	CMP	R6, R10			; pointer to data segment?
		BLT	rel_loop2
		SUB	R6, R6, R10
		ADD	R6, R6, R2
		B	rel_loop3
rel_loop2	SUB	R6, R6, R9
		ADD	R6, R6, R1
rel_loop3	STR	R6, [R1, R5]		; store it back
		SUBS	R4, R4, #1
		BGT	reloc_loop		; relocate next address

; now shift the data segment to the right place
move_dataseg
		LDR	R0, [R1, #FILE_DATA_LEN]; length of data segment
		LDR	R4, [R1, #FILE_TEXT_LEN]
		ADD	R4, R4, R1		; start of old data seg
		B	mv_dseg_test
mv_dseg_loop	LDR	R5, [R4, R0]
		STR	R5, [R2, R0]		; move to new address
mv_dseg_test	SUBS	R0, R0, #4
		BGE	mv_dseg_loop

; clear the bss segment
clr_bss
		LDR	R0, [R1, #FILE_CLR_BSS]
		MOV	R4, #&FB000000		; opcode for BLNV
		STR	R4, [R1, #FILE_CLR_BSS]	; disable second bss-clear
		AND	R4, R0, #&FF000000	; get opcode
		CMP	R4, #&EB000000		; is it Branch with Link (BL)?
		LDMNEIA	SP!, {R0-R5, LK}	; fetch R0 on error
		BNE	kernel_error		; no! But this must be!
		LDR	R0, [R1, #FILE_BSS_LEN]
		MOV	R4, #0			; section is to be cleared
		B	bss_loop_test
bss_loop	STR	R4, [R3, R0]		; clear one word
bss_loop_test	SUBS	R0, R0, #4
		BGE	bss_loop

; this part patches the list of addresses below into the first few bytes
; of the bss section of the program (only data of the first four parts)
patch_adr
		ADR	R4, kernel_text
		MOV	R0, #fsck_text-kernel_text
		B	patch_loop_test
patch_loop	LDR	R5, [R4, R0]		; copy one word
		STR	R5, [R3, R0]
patch_loop_test	SUBS	R0, R0, #4
		BGE	patch_loop

load_end	LDMIA	SP!, {R0-R5, LK}
		MOVS	PC, LK


;==============================================================================
; prints out load error; R0 points to the filename
;==============================================================================
kernel_error
		SPrint	"error loading "
		LDR	R0, act_name
		SWI	OS_Write0+XBit
		NLPrint
		SWI	OS_Exit



;==============================================================================
; data which are copied to high mem
;==============================================================================

path_var_name	=	"Minix$Path", 0
kernel_name	=	"kernel", 0
mm_name		=	"mm", 0
fs_name		=	"fs", 0
init_name	=	"init", 0
fsck_name	=	"fsck", 0
act_name	DCD	0		; pointer to name of actual module

kernel_text	DCD	0
kernel_data	DCD	0
kernel_bss	DCD	0
kernel_end	DCD	0
mm_text		DCD	0
mm_data		DCD	0
mm_bss		DCD	0
mm_end		DCD	0
fs_text		DCD	0
fs_data		DCD	0
fs_bss		DCD	0
fs_end		DCD	0
init_text	DCD	0
init_data	DCD	0
init_bss	DCD	0
init_end	DCD	0
fsck_text	DCD	0
fsck_data	DCD	0
fsck_bss	DCD	0
fsck_end	DCD	0

		%	&100
loader_stack	EQU	.
load_kernel_end	EQU	.

end		EQU	.		; Programmende

		END
