/* ---------------------------------------------------------------------------- *
 *	Kernel assembler library
 *
 *	Copyright (c) 2000, 20001 by G.Hopper
 * ---------------------------------------------------------------------------- */

/*	
#include <minix/config.h>
#include <minix/const.h>
#include "const.h"
#include "sconst.h"
#include "protect.h"
*/

/*	
! This file contains a number of assembly code utility routines needed by the
! kernel.  They are:
*/
/*	
!.global	_monitor	! exit Minix and return to the monitor
!.global	_int86		! let the monitor make an 8086 interrupt call
!.global	_cp_mess	! copies messages from source to destination
*/
.global	_exit		/* dummy for library routines	*/
.global	__exit		/* dummy for library routines	*/
.global	___exit		/* dummy for library routines	*/
.global	___main		/* dummy for GCC		*/
.global	_lock		/* disable interrupts		*/
.global	_unlock		/* enable interrupts		*/

/*	
!.global	_enable_irq	! enable an irq at the 8259 controller
!.global	_disable_irq	! disable an irq
!.global	_memcpy
*/
	
.global	_phys_copy		/* copy data from anywhere to anywhere in memory */

/*	
!.global	_mem_rdw	! copy one word from [segment:offset]
!.global	_reset		! reset the system
!.global	_mem_vid_copy	! copy data to video ram
!.global	_vid_vid_copy	! move data in video ram
!.global	_level0		! call a function at level 0
*/

/*
!*===========================================================================*
!*				phys_copy				     *
!*===========================================================================*
! PUBLIC void phys_copy(phys_bytes source, phys_bytes destination,
!			phys_bytes bytecount);
! Copy a block of physical memory.
!
! StrongARM phys_copy
!
! by UNO c4402 students (Golden boys)
!
! Entry: ?: destination pointer
!        ?: source pointer
!        ?: byte count
!
! Exit:  ?: destination pointer
!        ?: trashed
!
*/
_phys_copy:	      
        cmp	A3, #32			/*  more bytes to copy?			*/
	blt	pc_short_test		/* no					*/	

	tst	A1, #3			/* source: word aligned address?	*/	
	tsteq	A2, #3			/* destination: word aligned address?	*/
	bne	pc_short_test

	stmdb	SP!, {V1-V4}
	b	pc_long_test
pc_long_loop:
	ldmia	A1!, {V1-V4}		/* load 4 words				*/	
	stmia	A2!, {V1-V4}		/* store them				*/
pc_long_test:	
        subs	A3, A3, #16		/* another 16 Bytes to copy?		*/
	bge	pc_long_loop		/* yes!					*/
	ldmia	SP!, {V1-V4}
	adds	A3, A3, #16
	moveqs	PC, LR			/* no more single bytes to copy		*/	
	b	pc_short_test
pc_short_loop:	
        ldrb	A4, [A1, A3]
	strb	A4, [A2, A3]
pc_short_test:	
        subs	A3, A3, #1		/* another byte to copy?		*/	
	bge	pc_short_loop
	movs	PC, LR

/*
!*===========================================================================*
!*				lock					     *
!*===========================================================================*
; PUBLIC void lock();
; Disable CPU interrupts.
*/

_lock:	
	mov	R2, LR
	swi	MX_EnterSVC
	teqp	PC, #FIRQFlag:OR:IRQFlag:OR:SVCMode
	ldr	R0, _dis_irq_cnt
	adds	R0, R0, #1
	str	R0, _dis_irq_cnt
	movgts	PC, R2			/* return if already disabled */

	mov	R0, #IOC_BASE
	ldrb	R1, [R0, #IOC_IRQAMASK]
	strb	R1, _irq_mask1
	ldrb	R1, [R0, #IOC_IRQBMASK]
	strb	R1, _irq_mask2
	ldrb	R1, [R0, #IOC_FIRQMASK]
	strb	R1, _irq_mask3
	mov	R1, #0			/* disable all interrupts */
	strb	R1, [R0, #IOC_IRQAMASK]
	strb	R1, [R0, #IOC_IRQBMASK]
	strb	R1, [R0, #IOC_FIRQMASK]
	movs	PC, R2

_dis_irq_cnt:	DCD	-1
_irq_mask1:	%	1
_irq_mask2:	%	1
_irq_mask3:	%	1

/*	
!*===========================================================================*
!*				unlock					     *
!*===========================================================================*
; PUBLIC void unlock();
; Enable CPU interrupts.
;
*/

_unlock:	
	mov	R2, LR
	swi	MX_EnterSVC
	teqp	PC, #FIRQFlag:OR:IRQFlag:OR:SVCMode
	ldr	R0, dis_irq_cnt
	subs	R0, R0, #1
	str	R0, dis_irq_cnt
	movges	PC, R2
	mov	R0, #IOC_BASE
	ldrb	R1, irq_mask3
	strb	R1, [R0, #IOC_FIRQMASK]
	ldrb	R1, irq_mask2
	strb	R1, [R0, #IOC_IRQBMASK]
	ldrb	R1, irq_mask1
	strb	R1, [R0, #IOC_IRQAMASK]
        movs	PC, R2
	

/*				
!*===========================================================================*
!*				exit					     *
!*===========================================================================*
! PUBLIC void exit();
! Some library routines use exit, so provide a dummy version.
! Actual calls to exit cannot occur in the kernel.
! GNU CC likes to call ___main from main() for nonobvious reasons.
*/

_exit:
__exit:
___exit:

___main:
        
        .align	

